// -*- C++ -*-

// AudioIO.c
//
// Classes for sources and sinks of data
// that interface to real-time audio hardware
// for Audio Pattern Recognition Language (aprl)
// dpwe@media.mit.edu 1995feb02
//
#ifndef lint
static char rcsid[] = "$Header: /u/drspeech/src/aprl/RCS/AudioIO.cc,v 1.3 1999/01/09 02:29:41 dpwe Exp $ @media.mit.edu Machine Listening Group";
#endif /* lint */

#include <stdio.h>
#include <assert.h>
#include <math.h>	// for rint()

#include "AudioIO.h"

#define DATAFORMAT AFMT_FLOAT	/* if DATA==float, this must match */

/* #define REPORTDROPS	*/	/* if defined, will fprintf when dropped
                                   samples occur on input or outpur - but
				   the printing adds so much to the loop 
				   that you never recover */
void AudioSrc::init(void)
{   /* actually open the audio input */
    int 	format = DATAFORMAT;
    double	srate  = SampleRate();
//    int 	chans  = 1;	/* all we support for the moment */
    int 	desbuf = (int)(bufferTime * srate);
    int 	mode   = AIO_INPUT;

//    fprintf(stderr, "AudioSrc::init: chans %d, size %d, desb %d\n", 
//	    chans, Size(), desbuf);

    // request a buffer that is precisely double what we plan to use
    if( (au = AUOpen(format, srate, chans, 2*desbuf, mode)) == NULL)  {
	fprintf(stderr, "AudioSrc::init: fail on AUOpen(%d,%.0f,%d,%d,%d)\n", 
		format, srate, chans, 2*desbuf, mode);
	status = -1;	/* DON'T attempt to reopen */
    } else {
	/* else what? */
        status = 1;
    }
    lastTime = 0;
}

AudioSrc::~AudioSrc(void)
{   /* close the port before departing */
    auClose();
}

void AudioSrc::auOpen(void)
{   /* If the audio has been previously closed, open it again */
    int 	format = DATAFORMAT;
    double	srate  = SampleRate();
//    int 	chans  = 1;	/* all we support for the moment */
    int 	desbuf = (int)(bufferTime * srate);
    int 	mode   = AIO_INPUT;

    if (status != 1) {
        init();
    }
}

void AudioSrc::auClose(void)
{   /* If the audio is open, close it for now */
    if(au) {
	AUClose(au, 1);
	au = 0;
	status = -1;	/* paused - DON'T automatically reopen */
    }
}

#ifdef __sgi
#include <limits.h>	// for CLK_TCK on sgi
extern "C" int sginap(long ticks);
/* #define DO_SGINAP */
#endif /* __sgi */

int AudioSrc::Update(double time)
{   /* grab samples from the input */
    int totalGot = 0;
    if(status == 0)  {	// audio not yet started
	init();
    }
    int desSamps = min(Size(), (int)rint((time-lastTime)*SampleRate()));
    if(desSamps <= 0)		desSamps = Size()/chans; // of the circ. buffer
    lastTime = time;
    if(au) {
	int avaiSamps = (int)AUBufConts(au);
	if(avaiSamps < desSamps)  {
	    int mswait = (int)(1000.0*(desSamps-avaiSamps)/SampleRate());
#ifdef DO_SGINAP
	    if(mswait > (1000/CLK_TCK))  {
		sginap(mswait/(1000/CLK_TCK));	// release the CPU
	    }
#endif /* DO_SGINAP */
	}
	// OK, now if we block, that's just life
	int toget = desSamps*chans;
	DATA *d;
	int n;
	PushHead();
	double sc = scale;
	int doscale = (sc!=1.0 && sc!=0.0);
	while(toget)  {
	    n = min(toget, GetHeadPtr(&d));
	    AURead(au, d, n/chans);
	    if(doscale)   DATAscale(d, d, n, sc);
	    UpdateHead(n);
	    toget -= n;
	    totalGot += n;
	}
#ifdef REPORTDROPS
	int df;
	if((df=AUDroppedFrames(au))>0)  {
	    fprintf(stderr, "AudioSrc:%d frames dropped\n", df);
	}
#endif /* REPORTDROPS */
    }
    return totalGot;
}

// -------------- AudioSink

void AudioSink::init(void)
{   /* actually open the audio input */
    int 	format = DATAFORMAT;
    double	srate;
//    int 	chans  = 1;	/* all we support for the moment */
    int 	desbuf;
    int 	mode   = AIO_OUTPUT;

    if(Input())  {
	desbuf = Input()->Size();
	if(sampleRate==0)
	    srate  = Input()->SampleRate();
	else
	    srate  = sampleRate;
     } else {
	desbuf = (int)(DFLT_SRATE * DFLT_BUFTIME);
	srate  = DFLT_SRATE;
    }
//    fprintf(stderr, "AudioSink::init: chans %d, desb %d\n", 
//	    chans, desbuf);

    if( (au = AUOpen(format, srate, chans, 2*desbuf, mode)) == NULL)  {
	fprintf(stderr, "AudioSink::init: fail on AUOpen(%d,%.0f,%d,%d,%d)\n", 
		format, srate, chans, desbuf, mode);
    } else {
	/* else what? */
    }
}

AudioSink::~AudioSink(void)
{   /* close the port before departing */
    if(au)  {
	AUClose(au, 0);		/* allow buffer to empty before killing */
    }
}

int AudioSink::Update(double /* time */ )
{   /* grab samples from the input */
    int totalDone = 0;
    if(ThereIsMore())  {
	if(au==0)  {		// audio not yet started
	    init();
	}
	if(au)  {
	    assert(Input()->PopTail()>=0);
	    int todo  = Input()->Avai();
	    int avaiSamps = (int)AUBufSpace(au);
	    if(avaiSamps < todo)  {
		int mswait = (int)(1000.0*(todo-avaiSamps)/Input()->SampleRate());
#ifdef DO_SGINAP
		if(mswait > (1000/CLK_TCK))  {
		    sginap(mswait/(1000/CLK_TCK)); // release the CPU
		}
#endif	/* DO_SGINAP */
	    }
	    DATA *d;
	    int n;
	    double sc = scale, isc = (1.0/scale);
	    int doscale = (sc != 1.0 && sc != 0.0);
	    while(todo)  {
		n = min(todo, Input()->GetTailPtr(&d));
		if(doscale)   DATAscale(d, d, n, sc);
		AUWrite(au, d, n/chans);
		if(doscale)   DATAscale(d, d, n, isc);
		Input()->UpdateTail(n);
		todo -= n;
		totalDone += n;
	    }	
	    // inputCount += totalDone;
#ifdef REPORTDROPS
	    int df;
	    if((df=AUDroppedFrames(au))>0)  {
		fprintf(stderr, "AudioSink:%d frames dropped\n", df);
	    }
#endif /* REPORTDROPS */
	}
    }
    /* fprintf(stderr, "%d samps written\n", totalDone); */
    return totalDone;
}

#ifdef MAIN			// ------------- TEST CODE

extern "C" {
    char *programName;
};

#include "Sound.H"

void SaveAsSound(DataSource *cb, char *filename)
{				/* write the entire contents of the circular buffer out to disk 
				   as a soundfile. */
    Sound snd("w", cb->SampleRate(), 1, Sfmt_SHORT);
    snd.open(filename);
    snd.setFormat(Sfmt_FLOAT);
    int n, todo = cb->Size(), pos = cb->Count()-todo;
    DATA *d;
    while(todo)  {
	n = min(todo, cb->GetTailPtrFrom(&d, pos));
	snd.writeFrames(d,n);
	fprintf(stderr, "wrote %d\n", n);
	todo -= n;
	pos += n;
    }
}

void testAIO(double sr= 8000.0)
{
    AudioSrc  as(sr, 0.25);
    AudioSink ak(&as);
    
    SchedProc *sp = &as;
    fprintf(stderr, "as 0x%lx, sp 0x%lx\n", &as, sp);
    
    Scheduler s;
    
    s.Enschedule(sp);
    s.Enschedule(&ak);
    
    int loops = 1000;
    while(loops--)  {
	s.Update(0.0);
    }
    SaveAsSound(&as, "tmp.aif");
}

void main(int argc, char **argv)
{
    programName = argv[0];
    if(argc < 2)  {
	fprintf(stderr, "Usage: %s sampleRate\n", programName);
	exit(-1);
    }
    double sr = atof(argv[1]);
    testAIO(sr);
}

#endif				/* MAIN */
