// -*- C++ -*-

// DataCombiner.cc
//
// Class that takes several (two?) streams of data and combines them 
// to a single input.  In the first instance, we'll force them all to
// be at the same sampling rate, but eventually I'd like to support
// constants and maybe resampling as well.
// for Audio Pattern Recognition Language (aprl)
// dpwe@media.mit.edu 1995feb12
//

#include <stdio.h>
#include <assert.h>

inline double max(double a, double b)
{
    return (a>b)?a:b;
}

#include "DataCombiner.h"

void DataStreamCombiner::checkit(void)
{   
    DataSource *in1 = Input1();
    DataSource *in2 = Input2();
    if(in1!=0 || in2!=0)  {
	int maxsize = 0;
	double mxsq = 0;
	if(in1==0) { 
	    SetSampleRate(in2->SampleRate()); 
	    maxsize = in2->Size();
	    mxsq = in2->Sequence();
	} else if(in2==0) { 
	    SetSampleRate(in1->SampleRate()); 
	    maxsize = in1->Size();
	    mxsq = in1->Sequence();
	} else {  
	    double sr1 = in1->SampleRate();
	    double sr2 = in2->SampleRate();
	    if(sr1 != sr2)
		fprintf(stderr, "DataComb: srs of %.0f, %.0f diff\n", 
			sr1, sr2);
	    SetSampleRate(sr1);	// arbitrary choice 
	    maxsize = max(in1->Size(), in2->Size());
	    mxsq    = max(in1->Sequence(), in2->Sequence());
	}
	if(maxsize > Size())
	    Resize(maxsize);
	EnsureSeqAbove(mxsq);	// check sequencing OK for I/O
    } 	
}

int DataStreamCombiner::ThereIsMore(void)
{ 	
    // return sink1->ThereIsMore() && sink2->ThereIsMore(); 
    return DataSink1::ThereIsMore() 
	&& DataSink2::ThereIsMore();
}


int DataProcCombiner::Update(double /* time */)
{    /* actually invoke the process on the data. <time> ignored? */
    int totalDone=0;
    if(ThereIsMore())  {
	PushHead();		// subsequent PopTail()s on us will go here
	Input(1)->PopTail();	// reset to read from inputs
	Input(2)->PopTail();
	// weird circularbufferism
	int n1, n2, no, d1, d2, dout, done;
	int arediff = (Input(1) != Input(2));
	DATA *in1, *in2, *out;
	do  {
	    n1 = min(Input(1)->Avai(), Input(1)->GetTailPtr(&in1));
	    n2 = min(Input(2)->Avai(), Input(2)->GetTailPtr(&in2));
	    no = GetHeadPtr(&out);
	    if(n1>0 && n2>0 && no>0)  {
		d1 = n1; d2 = n2; dout = no;
		done = combFn->combine(in1, 1, &d1, 
				       in2, 1, &d2, 
				       out, 1, &dout);
		totalDone += done;
		Input(1)->UpdateTail(d1);
		// this is weird, but it makes it work if both inputs 
		// are in fact the same circular buffer (I hope)
		if(arediff)  {
		    Input(2)->UpdateTail(d2);
		}
		UpdateHead(dout);
	    } else {
		done = 0;
	    }
	} while(done);
    }
    return totalDone;
}

int DATAadder::combine(DATA *in1, int step1, int *ni1,
		       DATA *in2, int step2, int *ni2,
		       DATA *out, int stepo, int *no)
{   int todo = min(*no, min(*ni1, *ni2));
    int i;
    for(i = 0; i<todo; ++i)  {
	*out = k1 * *in1 + k2 * *in2;
	out += stepo;	in1 += step1; 	in2 += step2;
    }  
    *ni1 = todo; *ni2 = todo; *no = todo; return todo;  
}

int DATAmulter::combine(DATA *in1, int step1, int *ni1,
 		       DATA *in2, int step2, int *ni2,
		       DATA *out, int stepo, int *no)
{   int todo = min(*no, min(*ni1, *ni2));
    int i;
    for(i = 0; i<todo; ++i)  {
	*out = k1 * *in1 * k2 * *in2;
	out += stepo;	in1 += step1; 	in2 += step2;
    }  
    *ni1 = todo; *ni2 = todo; *no = todo; return todo;  
}


