// -*- C++ -*-

// MiscPipes.cc
//
// Classes for sources and sinks of data
// Various DataPipes.
// for Audio Pattern Recognition Language (aprl)
// dpwe@media.mit.edu 1995feb02
//
#ifndef lint
static char rcsid[] = "$Header: /u/drspeech/src/tcltk/aprl/RCS/MiscPipes.cc,v 1.2 1997/07/19 01:27:27 dpwe Exp $ @media.mit.edu Machine Listening Group";
#endif /* lint */

#ifndef HAVE_DRAND48
extern "C" double drand48(void);	// can't find it in /usr/include
#endif /* SUNOS41 */

#include <stdio.h>
#include <math.h>
#include <stdlib.h>	// for drand48 if nothing else
#include <assert.h>

#include "MiscPipes.h"

int RandSrc::Update(double time)
{   /* generate some random samples */
    int i, n, todo, totalDone = 0;
    DATA *d;

/*    todo = blksize;  */
    todo = (int)((time - lasttime)*SampleRate());
    lasttime = time;
//    fprintf(stderr, "RandSrc(%d):", todo);
    PushHead();
    while(todo) {
	assert(n = GetHeadPtr(&d));
	n = min(n, todo);
	for(i = 0; i<n; ++i)  {
	    d[i] = scale*drand48();
	    // fprintf(stderr, "%.3f ", d[i]);
	}
	UpdateHead(n);
	todo -= n;
	totalDone += n;
    }
//    fprintf(stderr, "\n");
//    CircularBuffer::print("RSrc");
    return totalDone;
}

int SquareProc::process(DATA *input, int istride, int niItems, 
			DATA *output, int ostride, int noItems, 
			int *poutdone)
{   /* square a block of data */
    int i,todo = min(niItems, noItems);
    DATA d;
    for(i=0; i<todo; ++i) { 
	d = *input; *output = d*d;
	input += istride; output += ostride; 
    }
    *poutdone = todo; 
    return todo; 
}

int RootProc::process(DATA *input, int istride, int niItems, 
		      DATA *output, int ostride, int noItems, 
		      int *poutdone)
{   /* square-root a block of data */
    int i,todo = min(niItems, noItems);
    DATA d;
    for(i=0; i<todo; ++i) { 
	d = *input; *output = sqrt(d);
	input += istride; output += ostride; 
    }
    *poutdone = todo; 
    return todo; 
}

int DummySink::Update(double /* time */)
{   /* just a place to leave data - prints it out */
    int totalDone = 0;
    if(ThereIsMore())  {
	Input()->PopTail();
	int avai = Input()->Avai();
	DATA *d;
	int i, n;
	fprintf(stderr, "DummySink(%d):", avai);
	while(totalDone < avai)  {
	    assert(n = Input()->GetTailPtr(&d));
	    n = min(n, avai-totalDone);
	    for(i = 0; i<n; ++i)  {
		fprintf(stderr, "%.3f ", d[i]);
	    }
	    Input()->UpdateTail(n);
	    // inputCount += n;
	    totalDone += n;
	}
	fprintf(stderr, "\n");
    }
    return totalDone;
}

RmsProc::RmsProc(int winlen, int step) : WinLen(winlen), Step(step)
{  
    store = new CircularBuffer(winlen); 
    win   = new CircularBuffer(winlen);
    // prefill window
    int i;  DATA *d;  win->PushHead(); win->GetHeadPtr(&d);
    for(i = 0; i<winlen; ++i)  *d++ = 0.5*(1-cos(i*2*PI/winlen));
    win->UpdateHead(winlen);
    // prefill storage buffer to half-full
    store->PushHead();
    store->GetHeadPtr(&d);  DATAclear(d, winlen/2);
    store->UpdateHead(winlen/2);
}

RmsProc::~RmsProc(void)	
{ 
    delete store; 
    delete win; 
}

int RmsProc::process(DATA *input, int istride, int niItems, 
		     DATA *output, int ostride, int noItems, 
		     int *poutdone)
{    
    int odone = 0; int idone = 0;
    int irem = niItems; int orem = noItems;
    while(irem && orem)  {
	int i, n, need = WinLen - store->Avai();
	DATA v, *d;
	n = store->GetHeadPtr(&d);
	int toget = min(need, min(n,irem));
	DATAstrideCopy(input, istride, d, 1, toget);
	for(i = 0; i<toget; ++i)  {
	    v = *input;  input += istride;
	    *d++ = v*v;
	}
	store->UpdateHead(toget);
	idone += toget;
	irem  -= toget;
	need  -= toget;
	if(need == 0)  {
	    store->PushTail();
	    win->PopTail();
	    *output = sqrt(CircBufInnerProd(win, store, WinLen)/(WinLen/2));
	    output += ostride;
	    ++odone; --orem;
	    store->PopTail();
	    store->UpdateTail(Step);
	}
    }
/*    fprintf(stderr, "RmsProc: i %d used %d, o %d used %d\n", 
	    niItems, idone, noItems, odone); */
    *poutdone = odone;
    return idone;
}

int DelayPipe::Update(double /* time */)
{   
    int totalDone = 0;
//    if(started==0)	fprintf(stderr, "Delay 0x%lx, size %d, avai %d\n", 
//			   this, delayBuffer->Size(), delayBuffer->Avai());
    started = 1;
    if(ThereIsMore())  {
	assert(delayBuffer->Size()==delayBuffer->Avai());
	PushHead();
	Input()->PopTail();
	int todo = Input()->Avai();
	if(todo > delayBuffer->Size())  {
	    // case 1 - copy out entire contents of delayBuffer
	    CircBufTransfer(delayBuffer, (CircularBuffer *)this, 
			    delayBuffer->Avai());
	    CircBufTransfer((CircularBuffer *)Input(), 
			    (CircularBuffer *)this, 
			    todo - delayBuffer->Size());
	    CircBufTransfer((CircularBuffer *)Input(), delayBuffer, 
			    delayBuffer->Size());
	    totalDone = todo;
	} else {		// transfer less than a whole buffer's worth
	    int nout = min(todo, delayBuffer->Avai());
	    CircBufTransfer(delayBuffer, (CircularBuffer *)this, nout);
	    CircBufTransfer((CircularBuffer *)Input(), delayBuffer, 
			    todo);
	    totalDone = todo;
	}
    }
    return totalDone; 
}
