// -*- C++ -*-

//
// Xing.h
//
// zero-crossing detector module for billg's pitch tracker
// based on billg's xing.h
//
// dpwe 1995feb24
//

typedef void (*VOID_FN_LONG_VP)(long, void *);

class xing {	// zero-crossing detector
private:
    // state
    int state;	    /* xing detector state */
    double env;	    /* envelope */
    long xing_c;	    /* count of last xing */
    long count;	    /* count since last accepted xing GLOBAL */

    int hpf_flag;	    /* TRUE if run highpass filter */
    double hpf_fc;	    /* cuttoff */
    double hpf_gain;     /* 1-pole highpass filter gain (zero at DC) */
    double hpf_z;	    /* delay element for hpf */

    int lpf_flag;	    /* TRUE if run lowpass filter */
    double lpf_fc;	    /* cuttoff */
    double lpf_gain;     /* 1-pole lowpass filter gain */
    double lpf_z;	    /* delay element for lpf */
    // parameters
    double decay_tc;     /* envelope decay time constant */
    double decay;	/* envelope decay */

    double hpf_proc(double samp);
    double lpf_proc(double samp);

protected:
    long samprate;

    // `accept fn' is pure virtual, called when crossing is accepted
    // delta is the time in samples since the last zero crossing.
    virtual void accept_fn(long delta) = 0;

public:
    xing(long srate);
    virtual ~xing(void);
    
    void default_params(void);
    void query_params(void);
    void proc(short samp);

    // accessors for 'global' vars
    int&    Hpf_flag(void)  { return hpf_flag; }; // TRUE if run highpass filt
    int&    Lpf_flag(void)  { return lpf_flag; }; // TRUE if run lowpass filt
    void    Set_hpf_fc(double f);
    void    Set_lpf_fc(double f);
    void    Set_decay_tc(double t);

};
