#!./aprlwish
#
# demo1.tcl
# 
# Show filtering of real-time input, display energy in different bands
#
# dpwe@media.mit.edu 1995feb14
#

#### Build the audio network

  # Define a scheduler object called 's'.  The scheduler must be the 
  # first object defined, since it controls the execution of all the 
  # other objects
scheduler s

  # Define an audio `source' (real-time input) called `as'
audiosrc  as    -sampleRate 11025
  # Define a band-pass filter, f1, centered at 5000 Hz, taking input from 
  # the most-recently-defined source
bpfilter  f1    -centerFreq 5000
  # Calculate the rms level of the signal, and downsample to 100 Hz
  # (the special name '.' means allocate a name arbitrarily)
rms       .     -sampleRate 100
  # Convert the linear rms valus to decibels (20log10), referenced so 
  # that 0 dB = 0.01
dB        .     -ref 0.01
  # Pass this output to a chart-recorder type display showing a range 
  # -20 to +20 (i.e. -1 to 1 after the scale factor 0.05 is applied)
wavesink  .ws1  -scale 0.05

  # Take the output from the filter and write it to the 
  # real-time audio output.  We specify the input explicitly rather 
  # than using the default (last-defined)
audiosink .     -input f1

  # Define a second filtered rms ticker-tape display around 200 Hz.
  # Again, we need to specify the input to the filter explicitly
  # else it would take input from the most recently declared object 
  # that has an output (the dB operator in this case)
bpfilter  f2    -centerFreq 200     -input as 
rms       .     -sampleRate 100
dB        .     -ref 0.01
wavesink  .ws2  -scale 0.05

#### Execution of the program

  # Make the displays appear in the main window
pack .ws1 .ws2

  # The audio network is only executed when we pass the 'step' 
  # command to the scheduler object 's', and then it calculates only 
  # one block of samples (by default, 0.1 seconds worth).
  # So, define a procedure that steps the scheduler, then arranges 
  # for itself to be run again after a short delay

proc run {} {
   s step
   after 10 {run}
}

  # Start the processing
run
