#!./aprlwish -f
#
# oscope.tcl
#
# Display a time waveform, oscilloscope-like, in aprl
#
# dpwe@media.mit.edu 1995mar14
#

set tcl_interactive 1
wm minsize . 1 1

scheduler s -docmd 1 
# -command {.tl config -maxVal %t -minVal [expr "%t-$disptime"] ; after 10 { %s step }}

# Configure the overall samplerate
set rate 8000
# How much time to display
set disptime 0.5
set dispsamps [expr "int($rate*$disptime)"]

audiosrc as -sampleRate $rate

wavesink .ws -buflen $dispsamps
timeline .tl -isVertical 0 -tixTopLeft 1

# function to bind to scale widget to adjust scale const in object
proc setSrcScale {src max scale} {
    set val [expr "$scale/$max"]
    .asf.val config -text $val
    $src config -scale $val
}

# from is top of scale, to is bottom
frame .asf -borderwidth 2 -relief groove
scale .asf.sc -to 0 -from 100  -showvalue 0 -command "setSrcScale as 10.0 "
label .asf.label -text "Gain"
label .asf.val -text "0"

proc setTimeScale {wsw max scale} {
    global rate disptime
    set scale [expr "int($scale - $max)"]
    set range [expr "$scale/3"]
    set pten  [expr "10.0*pow(10.0,$range)"]
    set arg   [lindex {1.0 2.0 5.0} [expr "$scale - 3*$range"]]
        
#    set disptime [expr "exp(($scale-$max)*0.767528)"]
    set disptime [expr "$arg * $pten"]
    set dispsamps [expr "int($rate*$disptime)"]
    $wsw config -buflen $dispsamps
    .tsf.val config -text $disptime
#    .tl config -minVal 0 -maxVal $disptime
    .tl config -minVal -$disptime -maxVal 0
}

# from is top of scale, to is bottom
frame .tsf -borderwidth 2 -relief groove
scale .tsf.ts -to 0 -from 10  -showvalue 0 -command "setTimeScale .ws 10.0 "
label .tsf.label -text "Time base"
label .tsf.val -text "0"

# buttons
frame .bf 
button .bf.pause -text "Pause" -command { if { [lindex [.bf.pause config -text] 4]=="Pause" } { .bf.pause config -text "Resume"; s config -docmd 0 } else { .bf.pause config -text "Pause"; s config -docmd 1; s step }}
button .bf.quit  -text "Quit" -command "exit"

pack .bf.pause .bf.quit -side left -expand y
pack .bf -side bottom -fill both -padx 10 -pady 10

pack .asf.label .tsf.label -side top -fill both -padx 5 -pady 5
pack .asf.val .tsf.val -side bottom -fill both -padx 5 -pady 5
pack .asf.sc .tsf.ts       -side top -fill both -expand y -padx 5 -pady 5
pack .asf .tsf -side left -fill both  -padx 5 -pady 5
pack .tl -side bottom -fill both
pack .ws -side top -fill both -expand y

.tsf.ts set 5
.asf.sc set 50

s step

