//
// tkNeedle.cc
//
// tk widget to display a needle
// eds@media.mit.edu 25 Feb 95
//

#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <math.h>

extern "C" {
#include <tcl.h>
#include <tk.h>
}

// Widget data structure

typedef struct {
    // standard fields
    Tk_Window	tkwin;
    Display	*display;
    Tcl_Interp	*interp;
    Tk_3DBorder bgBorder;	/* Used for drawing background. */
    XColor	*fgColor;
    int 	borderWidth;
    int 	padX;
    int 	padY;
    int 	width, height;
    int 	relief;		// appearance of whole window
    // data for the drawing activities
    GC  	gc;
    int 	doubleBuffer;	// do we double buffer?
    int 	updatePending;
    // special colors for the leds
    XColor	*clipColor;
    // fields describing the content
    // config
    int frameWidth;
    double val;
    double scale;
    double offs;
    double damping;
    double vel;
    double pos,clipval;
  } Needle;
  

extern "C" {
// static function prototypes

int Needle_Init(Tcl_Interp *interp);
    /* add the tcl commands defined in this package to the interpreter */
int NeedleCmd(ClientData clientData, Tcl_Interp *interp, 
	       int argc, char **argv);
static int  NeedleConfigure(Tcl_Interp *interp, Needle *sw, 
			     int argc, char **argv, int flags);
static void NeedleDestroy(char* clientData);
static void NeedleDisplay(ClientData clientData);
static void NeedleEventProc(ClientData clientData, XEvent *eventPtr);
static int  NeedleWidgetCmd(ClientData clientData, Tcl_Interp * interp, 
			     int argc, char **argv);
}

static Tk_ConfigSpec NeedleConfigSpecs[] = {
    {TK_CONFIG_BORDER, "-background", "background", "Background", 
     "#cdb79e", Tk_Offset(Needle, bgBorder), 0 }, 
    {TK_CONFIG_SYNONYM, "-bg", "background", NULL, NULL, 0, 0}, 
    {TK_CONFIG_COLOR, "-foreground", "foreground", "Foreground", 
     "black", Tk_Offset(Needle, fgColor), 0 }, 
    {TK_CONFIG_SYNONYM, "-fg", "foreground", NULL, NULL, 0, 0}, 
    {TK_CONFIG_PIXELS, "-borderwidth", "borderWidth", "BorderWidth",
	"2", Tk_Offset(Needle, borderWidth), 0},
    {TK_CONFIG_SYNONYM, "-bw", "borderWidth", NULL, NULL, 0, 0}, 
    {TK_CONFIG_PIXELS, "-padX", "padX", "PadX",
	"2", Tk_Offset(Needle, padX), 0},
    {TK_CONFIG_PIXELS, "-padY", "padY", "PadY",
	"2", Tk_Offset(Needle, padY), 0},
    {TK_CONFIG_PIXELS, "-width", "width", "Width",
	"400", Tk_Offset(Needle, width), 0},
    {TK_CONFIG_PIXELS, "-height", "height", "Height",
	"32", Tk_Offset(Needle, height), 0},
    {TK_CONFIG_RELIEF, "-relief", "relief", "Relief",
	"raised", Tk_Offset(Needle, relief), 0},
    {TK_CONFIG_INT, "-dbl", "doubleBuffer", "DoubleBuffer",
	"1", Tk_Offset(Needle, doubleBuffer), 0},
    {TK_CONFIG_INT, "-framewidth", "framewidth", "FrameWidth",
	"3", Tk_Offset(Needle, frameWidth), 0},
    {TK_CONFIG_DOUBLE, "-value", "value", "Value", 
        "0.0", Tk_Offset(Needle, val), 0},
    {TK_CONFIG_DOUBLE, "-scale", "scale", "Scale", 
        "1.0", Tk_Offset(Needle, scale), 0},
    {TK_CONFIG_DOUBLE, "-offset", "offset", "Offset", 
        "0.0", Tk_Offset(Needle, offs), 0},
    {TK_CONFIG_DOUBLE, "-damping", "damping", "Damping",
        "0.7", Tk_Offset(Needle, damping), 0},
    {TK_CONFIG_COLOR, "-clipcolor", "clipcolor", "ClipColor", 
        "yellow", Tk_Offset(Needle, clipColor), 0 },
    {TK_CONFIG_DOUBLE,"-clipval","clipval","ClipVal",
        "0.8", Tk_Offset(Needle, clipval), 0},
    {TK_CONFIG_END, NULL, NULL, NULL, NULL, 0, 0}
};

int Needle_Init(Tcl_Interp *interp)
{   /* add the tcl commands defined in this package to the interpreter */
    Tk_Window main = Tk_MainWindow(interp);

    Tcl_CreateCommand(interp, "needle", NeedleCmd, (ClientData) main,
	    (void (*)(ClientData)) NULL);
    return TCL_OK;
}

int NeedleCmd(ClientData clientData, Tcl_Interp *interp, 
	       int argc, char **argv)
{   /* Tcl command to instance a new Needle widget */
    Tk_Window	main = (Tk_Window) clientData; 
    Needle 	*tmln; 
    Tk_Window	tkwin;

    if(argc < 2)  {
	Tcl_AppendResult(interp, "wrong # args: should be\"", 
			 argv[0], " pathName ?options?\"", NULL);
	return TCL_ERROR;
    }
    tkwin = Tk_CreateWindowFromPath(interp, main, argv[1], NULL);
    if(tkwin == NULL)  {
	return TCL_ERROR;
    }
    Tk_SetClass(tkwin, "Needle");

    // Allocate and initialize the structure
    tmln = (Needle *)ckalloc(sizeof(Needle));
    tmln->tkwin   = tkwin;
    tmln->display = Tk_Display(tkwin);
    tmln->interp  = interp;
    tmln->bgBorder= NULL;
    tmln->fgColor = NULL;
    tmln->clipColor = NULL;
    tmln->gc      = None;
    tmln->doubleBuffer = 1;
    tmln->updatePending = 0;

    tmln->val = 0.0;
    tmln->vel = 0.0;
    tmln->pos = 0.0;
    tmln->clipval = 0.0;

    Tk_CreateEventHandler(tmln->tkwin, ExposureMask|StructureNotifyMask,
			  NeedleEventProc, (ClientData)tmln);
    Tcl_CreateCommand(interp, Tk_PathName(tmln->tkwin), NeedleWidgetCmd, 
		      (ClientData)tmln, (void (*)(ClientData))NULL);
    if(NeedleConfigure(interp, tmln, argc-2, argv+2, 0) != TCL_OK)  {
	Tk_DestroyWindow(tmln->tkwin);
	return TCL_ERROR;
    }
    interp->result = Tk_PathName(tmln->tkwin);
    return TCL_OK;
}

static void NeedleGetWorkArea(Needle *tmln, 
				int *px, int *py, int *pw, int *ph)
{   /* return the area that time line legend is actually drawn in */
    Tk_Window tkwin = tmln->tkwin;

    *px = tmln->borderWidth + tmln->padX;
    *pw = Tk_Width(tkwin) - 2*(tmln->padX+tmln->borderWidth);
    *py = tmln->borderWidth + tmln->padY;
    *ph = Tk_Height(tkwin) - 2*(tmln->padY+tmln->borderWidth);
}

static int NeedleWidgetCmd(ClientData clientData, Tcl_Interp *interp, 
			    int argc, char **argv)
{   /* passed commands executed on the widget */
    Needle	*tmln = (Needle *)clientData;
    int 	result = TCL_OK;
    int 	length;
    char	c;

    if(argc < 2)  {
	Tcl_AppendResult(interp, "wrong # args: should be \"", 
			 argv[0], " option ?arg arg ...?\"", NULL);
	return TCL_ERROR;
    }
    Tk_Preserve((ClientData)tmln);
    c = argv[1][0];
    length = strlen(argv[1]);
    if( (c=='c') && (strncmp(argv[1], "configure", length)==0) )  {
	if(argc==2) {
	    result =Tk_ConfigureInfo(interp, tmln->tkwin, NeedleConfigSpecs, 
				      (char *)tmln, NULL, 0);
	} else if(argc==3) {
	    result =Tk_ConfigureInfo(interp, tmln->tkwin, NeedleConfigSpecs, 
				      (char *)tmln, argv[2], 0);
	} else {
	    result =NeedleConfigure(interp, tmln, argc-2, argv+2, 
				      TK_CONFIG_ARGV_ONLY);
	}
    } else if ((c == 'c') && (strncmp(argv[1], "cget", length) == 0)) {
	if(argc != 3)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " cget option\"", (char *)NULL);
	    return TCL_ERROR;
	}
	result = Tk_ConfigureValue(interp, tmln->tkwin, NeedleConfigSpecs,
				   (char *) tmln, argv[2], 0);
    } else {
	Tcl_AppendResult(interp, "bad option \"", argv[1],
			 "\":  must be configure, cget", 
			 (char *) NULL);
	result = TCL_ERROR;
    }
    if (result == TCL_OK && !tmln->updatePending) {
	Tk_DoWhenIdle(NeedleDisplay, (ClientData) tmln);
	tmln->updatePending = 1;
    }
    Tk_Release((ClientData)tmln);
    return result;
}

// from tclAprlMisc.h
extern int ConfigSpecified(char *val, int argc, char **argv);
    // Checks config args for a particular specifier

static int NeedleConfigure(Tcl_Interp *interp, Needle *tmln, 
			    int argc, char **argv, int flags)
{   /* handle configure commands */
    int rslt = TCL_OK;

    if(Tk_ConfigureWidget(interp, tmln->tkwin, NeedleConfigSpecs, 
			  argc, argv, (char *)tmln, flags)!= TCL_OK)  {
	return TCL_ERROR;
    }
    /* Convert orientation string to actual direction */

    /* ensure values are OK */
    // set the colors and GC for the redisplay
    if(tmln->gc == None 	// first time through
       || ConfigSpecified("-background", argc, argv)	
       || ConfigSpecified("-bg", argc, argv)) {
	Tk_SetWindowBackground(tmln->tkwin, 
			       Tk_3DBorderColor(tmln->bgBorder)->pixel);
    }
    if (tmln->gc == None) {
	XGCValues gcValues;
	gcValues.function = GXcopy;
	gcValues.graphics_exposures = False;
	tmln->gc = Tk_GetGC(tmln->tkwin,GCFunction|GCGraphicsExposures, 
			    &gcValues);
    }
    // register desired window geometry, arrange for redisplay
    Tk_GeometryRequest(tmln->tkwin, tmln->width, tmln->height);
    Tk_SetInternalBorder(tmln->tkwin, tmln->borderWidth);
    if (!tmln->updatePending) {
	Tk_DoWhenIdle(NeedleDisplay, (ClientData)tmln);
	tmln->updatePending = 1;
    }
    return rslt;
}

static void NeedleEventProc(ClientData clientData, XEvent *eventPtr)
{				/* handle events for the Needle widget */
    Needle *tmln = (Needle *)clientData;
    
    if(eventPtr->type == Expose)  {
	if(!tmln->updatePending)  {
	    Tk_DoWhenIdle(NeedleDisplay, (ClientData)tmln);
	    tmln->updatePending = 1;
	}
    } else if(eventPtr->type == ConfigureNotify)  {
	// adjust self for resize 
	if(!tmln->updatePending)  {
	    Tk_DoWhenIdle(NeedleDisplay, (ClientData)tmln);
	    tmln->updatePending = 1;
	}
    } else if(eventPtr->type == DestroyNotify)  {
	Tcl_DeleteCommand(tmln->interp, Tk_PathName(tmln->tkwin));
	tmln->tkwin = NULL;
	if(tmln->updatePending)  {
	    Tk_CancelIdleCall(NeedleDisplay, (ClientData)tmln);
	}
	Tcl_EventuallyFree((ClientData)tmln, (Tcl_FreeProc *)NeedleDestroy);
    }
}

#define MAX(a,b) (((a)>(b))?(a):(b))
#define MIN(a,b) (((a)<(b))?(a):(b))
#define PI 3.14159265358979323

static void NeedleDraw(Needle *n, Drawable d=None)
{
  Tk_Window    tkwin = n->tkwin;
  GC           gc    = n->gc;
  Display      *disp = Tk_Display(tkwin);

  int x, y, w, h;
  double r,cx,cy,angle,angled,i,ccx,ccy;


  if (d == None)
    d = Tk_WindowId(tkwin);

  XSetForeground(disp,gc,n->fgColor->pixel);

  NeedleGetWorkArea(n,&x,&y,&w,&h);

  angle = PI/2;
  angled = angle / PI * 180;
  r = sqrt((.7 * w)*(.7 * w) / (1 - cos(angle)));
  if (r > h * 1.5) r = 1.5 * h;
  cx = x + w/2;
  cy = y + h + r/2;

  // physical model of the meter
  n->vel = n->vel + ((n->val-n->offs)*n->scale - n->pos);
  if (n->vel < 0)
    n->vel = n->vel * n->damping * n->damping;
  else
    n->vel = n->vel * n->damping;
  n->pos = n->pos + n->vel;
  
  if (n->pos < 0) n->pos = n->vel = 0;
  if (n->pos > 1) { n->pos = 1; n->vel = 0; }
  
  XDrawArc(disp,d,gc,(int)(cx-r),(int)(cy-r),(int)(2*r),(int)(2*r),
	  (int)((angled/2+90) * 64),-(int)(angled * 64));
  XDrawArc(disp,d,gc,(int)(cx-r*.98),(int)(cy-r*.98),
	   (int)(1.96*r),(int)(1.96*r),
	   (int)((PI/2 * 5/10 + angle/2) / PI * 180 * 64),
	   (int)((PI/2 * -5/10) / PI * 180 * 64));
  XDrawArc(disp,d,gc,(int)(cx-r*.96),(int)(cy-r*.96),
	   (int)(1.92*r),(int)(1.92*r),
	   (int)((PI/2 * 2/10 + angle/2) / PI * 180 * 64),
	   (int)((PI/2 * -2/10) / PI * 180 * 64));
	   
  for (i=0;i<=10.5;i+=1)
    XDrawLine(disp,d,gc,
	    (int)(cx + r * cos(PI/2 *(1-i/10) + angle/2)),
	    (int)(cy - r * sin(PI/2 *(1-i/10) + angle/2)),
	    (int)(cx + r*.95 * cos(PI/2 *(1-i/10) + angle/2)),
	    (int)(cy - r*.95 * sin(PI/2 *(1-i/10) + angle/2)));	
	   
  XDrawLine(disp,d,gc,
	    (int)(cx + r/2 * cos(PI/2 *(1-(n->pos)) + angle/2)),
	    (int)(cy - r/2 * sin(PI/2 *(1-(n->pos)) + angle/2)),
	    (int)(cx + r*.9 * cos(PI/2 *(1-(n->pos)) + angle/2)),
	    (int)(cy - r*.9 * sin(PI/2 *(1-(n->pos)) + angle/2)));

  ccx = (int)(w/2 - r/20);
  ccy = (int)(h*.95 - r/20);

  XFillArc(disp,d,gc,(int)ccx,(int)ccy,(int)(r/10),(int)(r/10),0,360*64);

  if (n->pos > n->clipval) { // clip
    ccx = (int)(w/2 - r/24);
    ccy = (int)(h*.95 - r/24);
    XSetForeground(disp,gc,n->clipColor->pixel);     
    XFillArc(disp,d,gc,(int)ccx,(int)ccy,(int)(r/12),(int)(r/12),0,360*64);
  }

    
}

static void NeedleDisplay(ClientData clientData)
{
  Needle *n = (Needle *)clientData;
  Tk_Window tkwin = n->tkwin;
  Pixmap pm = None;
  Drawable d;
  Display *disp = n->display;
  int tw,th,inset;

  n->updatePending = 0;
  if (!Tk_IsMapped(tkwin)) { return; }

  tw = Tk_Width(tkwin);
  th = Tk_Height(tkwin);

  if (n->doubleBuffer) {
    pm = XCreatePixmap(disp, Tk_WindowId(tkwin), tw, th, 
			   DefaultDepthOfScreen(Tk_Screen(tkwin)));
    d = pm;
  } else {
    d = Tk_WindowId(tkwin);
  }
  Tk_Fill3DRectangle(tkwin,d,n->bgBorder,0,0,tw,th,n->borderWidth,
		     n->relief);
  inset = 2;
  NeedleDraw(n,d);
  if (n->doubleBuffer) {
    XCopyArea(disp,pm,Tk_WindowId(tkwin),n->gc,0,0,tw,th,0,0);
    XFreePixmap(disp,pm);
  }
}

static void NeedleDestroy(char* clientData)
{				/* destroy one of our objects */
    Needle *n = (Needle *)clientData;
    
    Tk_FreeOptions(NeedleConfigSpecs, (char *)n, n->display, 0);
    if(n->gc != None)  {
	Tk_FreeGC(n->display, n->gc);
    }
    // don't destroy sound as we didn't new it
    ckfree((char *)n);
}
