// -------------------------------------------------------------------
// tkVUMeter.cc
//
// tk widget to display a VUMeter label
// dpwe@mit.edu 1994sep14
//

#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <math.h>

extern "C" {
#include <tcl.h>
#include <tk.h>
}

// Widget data structure

typedef struct {
    // standard fields
    Tk_Window	tkwin;
    Display	*display;
    Tcl_Interp	*interp;
    Tk_3DBorder bgBorder;	/* Used for drawing background. */
    XColor	*fgColor;
    int 	borderWidth;
    int 	padX;
    int 	padY;
    int 	width, height;
    int 	relief;		// appearance of whole window
    // data for the drawing activities
    GC  	gc;
    int 	doubleBuffer;	// do we double buffer?
    int 	updatePending;
    // special colors for the leds
    XColor	*greenColor;
    XColor	*yellowColor;
    XColor	*redColor;
    // fields describing the content
    // config
    Tk_Uid orientUid;		/* Orientation for window ("vertical" or
				 * "horizontal"). */
    int vertical;		/* Non-zero means vertical orientation
				 * requested, zero means horizontal. */
    int frameWidth;
    int		ncells;		// total number of VU meter elements
    int		ngreen;		// the bottom few are green
    int		nyellow;	// .. then yellow, rest are red
    // actual val
    double 	val;		// what proportion of cells are lit
    double	scale;		// mult val by this
    double 	offs;		// and add this, view 0..1
} VUMeter;

extern "C" {
// static function prototypes

int VUMeter_Init(Tcl_Interp *interp);
    /* add the tcl commands defined in this package to the interpreter */
int VUMeterCmd(ClientData clientData, Tcl_Interp *interp, 
	       int argc, char **argv);
static int  VUMeterConfigure(Tcl_Interp *interp, VUMeter *sw, 
			     int argc, char **argv, int flags);
static void VUMeterDestroy(char* clientData);
static void VUMeterDisplay(ClientData clientData);
static void VUMeterEventProc(ClientData clientData, XEvent *eventPtr);
static int  VUMeterWidgetCmd(ClientData clientData, Tcl_Interp * interp, 
			     int argc, char **argv);
}

// Argv parsing info

static Tk_ConfigSpec VUMeterConfigSpecs[] = {
    {TK_CONFIG_BORDER, "-background", "background", "Background", 
     "#cdb79e", Tk_Offset(VUMeter, bgBorder), 0 }, 
    {TK_CONFIG_SYNONYM, "-bg", "background", NULL, NULL, 0, 0}, 
    {TK_CONFIG_COLOR, "-foreground", "foreground", "Foreground", 
     "black", Tk_Offset(VUMeter, fgColor), 0 }, 
    {TK_CONFIG_SYNONYM, "-fg", "foreground", NULL, NULL, 0, 0}, 
    {TK_CONFIG_PIXELS, "-borderwidth", "borderWidth", "BorderWidth",
	"2", Tk_Offset(VUMeter, borderWidth), 0},
    {TK_CONFIG_SYNONYM, "-bw", "borderWidth", NULL, NULL, 0, 0}, 
    {TK_CONFIG_PIXELS, "-padX", "padX", "PadX",
	"2", Tk_Offset(VUMeter, padX), 0},
    {TK_CONFIG_PIXELS, "-padY", "padY", "PadY",
	"2", Tk_Offset(VUMeter, padY), 0},
    {TK_CONFIG_PIXELS, "-width", "width", "Width",
	"400", Tk_Offset(VUMeter, width), 0},
    {TK_CONFIG_PIXELS, "-height", "height", "Height",
	"32", Tk_Offset(VUMeter, height), 0},
    {TK_CONFIG_RELIEF, "-relief", "relief", "Relief",
	"raised", Tk_Offset(VUMeter, relief), 0},
    {TK_CONFIG_INT, "-dbl", "doubleBuffer", "DoubleBuffer",
	"1", Tk_Offset(VUMeter, doubleBuffer), 0},
    {TK_CONFIG_UID, "-orient", "orient", "Orient",
	"horizontal", Tk_Offset(VUMeter, orientUid), 0},
    {TK_CONFIG_COLOR, "-redcolor", "redcolor", "RedColor", 
        "red", Tk_Offset(VUMeter, redColor), 0 }, 
    {TK_CONFIG_COLOR, "-yellowcolor", "yellowcolor", "YellowColor", 
        "yellow", Tk_Offset(VUMeter, yellowColor), 0 }, 
    {TK_CONFIG_COLOR, "-greencolor", "greencolor", "GreenColor", 
        "green", Tk_Offset(VUMeter, greenColor), 0 }, 
    {TK_CONFIG_INT, "-framewidth", "framewidth", "FrameWidth",
	"3", Tk_Offset(VUMeter, frameWidth), 0},
    {TK_CONFIG_INT, "-cells", "cells", "Cells",
	"20", Tk_Offset(VUMeter, ncells), 0},
    {TK_CONFIG_INT, "-ngreen", "ngreen", "NGreen",
	"10", Tk_Offset(VUMeter, ngreen), 0},
    {TK_CONFIG_INT, "-nyellow", "nyellow", "nyellow",
	"5", Tk_Offset(VUMeter, nyellow), 0},
    {TK_CONFIG_DOUBLE, "-value", "value", "Value", 
        "0.0", Tk_Offset(VUMeter, val), 0},
    {TK_CONFIG_DOUBLE, "-scale", "scale", "Scale", 
        "1.0", Tk_Offset(VUMeter, scale), 0},
    {TK_CONFIG_DOUBLE, "-offset", "offset", "Offset", 
        "0.0", Tk_Offset(VUMeter, offs), 0},
    {TK_CONFIG_END, NULL, NULL, NULL, NULL, 0, 0}
};

int VUMeter_Init(Tcl_Interp *interp)
{   /* add the tcl commands defined in this package to the interpreter */
    Tk_Window main = Tk_MainWindow(interp);

    Tcl_CreateCommand(interp, "vumeter", VUMeterCmd, (ClientData) main,
	    (void (*)(ClientData)) NULL);
    return TCL_OK;
}

int VUMeterCmd(ClientData clientData, Tcl_Interp *interp, 
	       int argc, char **argv)
{   /* Tcl command to instance a new VUMeter widget */
    Tk_Window	main = (Tk_Window) clientData; 
    VUMeter 	*tmln; 
    Tk_Window	tkwin;

    if(argc < 2)  {
	Tcl_AppendResult(interp, "wrong # args: should be\"", 
			 argv[0], " pathName ?options?\"", NULL);
	return TCL_ERROR;
    }
    tkwin = Tk_CreateWindowFromPath(interp, main, argv[1], NULL);
    if(tkwin == NULL)  {
	return TCL_ERROR;
    }
    Tk_SetClass(tkwin, "VUMeter");

    // Allocate and initialize the structure
    tmln = (VUMeter *)ckalloc(sizeof(VUMeter));
    tmln->tkwin   = tkwin;
    tmln->display = Tk_Display(tkwin);
    tmln->interp  = interp;
    tmln->bgBorder= NULL;
    tmln->fgColor = NULL;
    tmln->gc      = None;
    tmln->doubleBuffer = 1;
    tmln->updatePending = 0;

    tmln->greenColor = NULL;
    tmln->yellowColor = NULL;
    tmln->redColor = NULL;
    tmln->ncells = 0;
    tmln->ngreen = 0;
    tmln->nyellow = 0;
    tmln->val = 0.0;

    Tk_CreateEventHandler(tmln->tkwin, ExposureMask|StructureNotifyMask,
			  VUMeterEventProc, (ClientData)tmln);
    Tcl_CreateCommand(interp, Tk_PathName(tmln->tkwin), VUMeterWidgetCmd, 
		      (ClientData)tmln, (void (*)(ClientData))NULL);
    if(VUMeterConfigure(interp, tmln, argc-2, argv+2, 0) != TCL_OK)  {
	Tk_DestroyWindow(tmln->tkwin);
	return TCL_ERROR;
    }
    interp->result = Tk_PathName(tmln->tkwin);
    return TCL_OK;
}

static void VUMeterGetWorkArea(VUMeter *tmln, 
				int *px, int *py, int *pw, int *ph)
{   /* return the area that time line legend is actually drawn in */
    Tk_Window tkwin = tmln->tkwin;

    *px = tmln->borderWidth + tmln->padX;
    *pw = Tk_Width(tkwin) - 2*(tmln->padX+tmln->borderWidth);
    *py = tmln->borderWidth + tmln->padY;
    *ph = Tk_Height(tkwin) - 2*(tmln->padY+tmln->borderWidth);
}

static int VUMeterWidgetCmd(ClientData clientData, Tcl_Interp *interp, 
			    int argc, char **argv)
{   /* passed commands executed on the widget */
    VUMeter	*tmln = (VUMeter *)clientData;
    int 	result = TCL_OK;
    int 	length;
    char	c;

    if(argc < 2)  {
	Tcl_AppendResult(interp, "wrong # args: should be \"", 
			 argv[0], " option ?arg arg ...?\"", NULL);
	return TCL_ERROR;
    }
    Tk_Preserve((ClientData)tmln);
    c = argv[1][0];
    length = strlen(argv[1]);
    if( (c=='c') && (strncmp(argv[1], "configure", length)==0) )  {
	if(argc==2) {
	    result =Tk_ConfigureInfo(interp, tmln->tkwin, VUMeterConfigSpecs, 
				      (char *)tmln, NULL, 0);
	} else if(argc==3) {
	    result =Tk_ConfigureInfo(interp, tmln->tkwin, VUMeterConfigSpecs, 
				      (char *)tmln, argv[2], 0);
	} else {
	    result =VUMeterConfigure(interp, tmln, argc-2, argv+2, 
				      TK_CONFIG_ARGV_ONLY);
	}
    } else if ((c == 'c') && (strncmp(argv[1], "cget", length) == 0)) {
	if(argc != 3)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " cget option\"", (char *)NULL);
	    return TCL_ERROR;
	}
	result = Tk_ConfigureValue(interp, tmln->tkwin, VUMeterConfigSpecs,
				   (char *) tmln, argv[2], 0);
    } else {
	Tcl_AppendResult(interp, "bad option \"", argv[1],
			 "\":  must be configure, cget", 
			 (char *) NULL);
	result = TCL_ERROR;
    }
    if (result == TCL_OK && !tmln->updatePending) {
	Tk_DoWhenIdle(VUMeterDisplay, (ClientData) tmln);
	tmln->updatePending = 1;
    }
    Tk_Release((ClientData)tmln);
    return result;
}

// from tclAprlMisc.h
extern int ConfigSpecified(char *val, int argc, char **argv);
    // Checks config args for a particular specifier

static int VUMeterConfigure(Tcl_Interp *interp, VUMeter *tmln, 
			    int argc, char **argv, int flags)
{   /* handle configure commands */
    int rslt = TCL_OK;

    if(Tk_ConfigureWidget(interp, tmln->tkwin, VUMeterConfigSpecs, 
			  argc, argv, (char *)tmln, flags)!= TCL_OK)  {
	return TCL_ERROR;
    }
    /* Convert orientation string to actual direction */
    int length = strlen(tmln->orientUid);
    if (strncmp(tmln->orientUid, "vertical", length) == 0) {
	tmln->vertical = 1;
    } else if (strncmp(tmln->orientUid, "horizontal", length) == 0) {
	tmln->vertical = 0;
    } else {
	Tcl_AppendResult(interp, "bad orientation \"", tmln->orientUid,
		"\": must be vertical or horizontal", (char *) NULL);
	return TCL_ERROR;
    }
    /* ensure values are OK */
    // set the colors and GC for the redisplay
    if(tmln->gc == None 	// first time through
       || ConfigSpecified("-background", argc, argv)	
       || ConfigSpecified("-bg", argc, argv)) {
	Tk_SetWindowBackground(tmln->tkwin, 
			       Tk_3DBorderColor(tmln->bgBorder)->pixel);
    }
    if (tmln->gc == None) {
	XGCValues gcValues;
	gcValues.function = GXcopy;
	gcValues.graphics_exposures = False;
	tmln->gc = Tk_GetGC(tmln->tkwin,GCFunction|GCGraphicsExposures, 
			    &gcValues);
    }
    // register desired window geometry, arrange for redisplay
    Tk_GeometryRequest(tmln->tkwin, tmln->width, tmln->height);
    Tk_SetInternalBorder(tmln->tkwin, tmln->borderWidth);
    if (!tmln->updatePending) {
	Tk_DoWhenIdle(VUMeterDisplay, (ClientData)tmln);
	tmln->updatePending = 1;
    }
    return rslt;
}

static void VUMeterEventProc(ClientData clientData, XEvent *eventPtr)
{				/* handle events for the VUMeter widget */
    VUMeter *tmln = (VUMeter *)clientData;
    
    if(eventPtr->type == Expose)  {
	if(!tmln->updatePending)  {
	    Tk_DoWhenIdle(VUMeterDisplay, (ClientData)tmln);
	    tmln->updatePending = 1;
	}
    } else if(eventPtr->type == ConfigureNotify)  {
	// adjust self for resize 
	if(!tmln->updatePending)  {
	    Tk_DoWhenIdle(VUMeterDisplay, (ClientData)tmln);
	    tmln->updatePending = 1;
	}
    } else if(eventPtr->type == DestroyNotify)  {
	Tcl_DeleteCommand(tmln->interp, Tk_PathName(tmln->tkwin));
	tmln->tkwin = NULL;
	if(tmln->updatePending)  {
	    Tk_CancelIdleCall(VUMeterDisplay, (ClientData)tmln);
	}
	Tcl_EventuallyFree((ClientData)tmln, (Tcl_FreeProc *)VUMeterDestroy);
    }
}

static void drawX(Tk_Window tkwin, GC gc, Drawable d=None, int inset=0)
{				/* write an X into a tkwin */
    Display *disp = Tk_Display(tkwin);
    int x, y, w, h;
    
    if(d==None)
	d = Tk_WindowId(tkwin);
    
    x = inset;
    w = Tk_Width(tkwin) - 2*x;
    y = inset;
    h = Tk_Height(tkwin) - 2*y;
    
    XDrawLine(disp, d, gc, x, y, x+w, y+h);
    XDrawLine(disp, d, gc, x, y+h, x+w, y);
}
#define MAX(a,b) (((a)>(b))?(a):(b))
#define MIN(a,b) (((a)<(b))?(a):(b))
#define MAXRECTS 64
static void VUMeterDraw(VUMeter *tmln, Drawable d=None)
{   /* actually draw the VUMeter data */
    Tk_Window 		tkwin = tmln->tkwin;
    GC 			gc    = tmln->gc;
    Display 		*disp = Tk_Display(tkwin);
    XColor 		*fgColor = tmln->fgColor;
    int x, y, w, h, v;
    XRectangle rect, rects[MAXRECTS];
    int totrs, thistime, i;
    int border = tmln->frameWidth;

    if(d==None)
	d = Tk_WindowId(tkwin);
    
    XSetForeground(disp, gc, tmln->fgColor->pixel);

    if(tmln->ncells <= 0)  {
	drawX(tkwin, gc, d, tmln->borderWidth+MAX(tmln->padX,tmln->padY));
	return;
    }

    VUMeterGetWorkArea(tmln, &x, &y, &w, &h);
    rect.x = x; rect.y = y; rect.width = w; rect.height = h;
    int drawcells = (int)(((tmln->scale*tmln->val)+tmln->offs)*tmln->ncells);
    if(drawcells < 0)	drawcells = 0;
    if(drawcells > tmln->ncells)	drawcells = tmln->ncells;
    // draw the background if fgColor
    XFillRectangle(disp, d, gc, x, y, w, h);
    // draw each of the green cells
    totrs = 0;
    while(totrs < tmln->ncells && totrs < drawcells)  {
	/* fprintf(stderr, "totrs %d, ng %d, ny %d, nc %d\n", 
		totrs, tmln->ngreen, tmln->nyellow, tmln->ncells);
	 */
	if(totrs >= (tmln->ngreen + tmln->nyellow))  {
	    fgColor = tmln->redColor;
	    thistime = MIN(drawcells, tmln->ncells) - totrs;
	} else if(totrs >= tmln->ngreen)  {
	    fgColor = tmln->yellowColor;
	    thistime = MIN(drawcells, tmln->ngreen+tmln->nyellow) - totrs;
	} else {
	    fgColor = tmln->greenColor;
	    thistime = MIN(drawcells, tmln->ngreen) - totrs;
	}
	thistime = MIN(thistime, MAXRECTS);
	for(i=0; i<thistime; ++i)  {
	    // set up rectangle
	    if(tmln->vertical)  {
		rects[i].x = x+border;	rects[i].width = w-2*border;
		rects[i].y = y+h+
		           -(int)((h-border)*(totrs+i+1)/(tmln->ncells));
		v = 
		        (y+h-border-(int)((h-border)*(totrs+i)/(tmln->ncells)))
			    - rects[i].y;
		rects[i].height = MAX(1, v);
	    } else {	// horizontal
		rects[i].y = y+border;	rects[i].height = h-2*border;
		rects[i].x = x+border+
		           (int)((w-border)*(totrs+i)/(tmln->ncells));
		v = 
		    (x+(int)((w-border)*(totrs+i+1)/(tmln->ncells)))
			- rects[i].x;
		rects[i].width = MAX(1, v);
	    }
	}
	XSetForeground(disp, gc, fgColor->pixel);
	XFillRectangles(disp, d, gc, rects, thistime);
	totrs += thistime;	
    }
}    	

static void VUMeterDisplay(ClientData clientData)
{				/* actually draw the contents of the widget */
    VUMeter *tmln = (VUMeter *)clientData;
    Tk_Window tkwin = tmln->tkwin;
    Pixmap pm = None;
    Drawable d;
    Display *disp = tmln->display;
    int 	tw, th, inset;
    
    tmln->updatePending = 0;
    if(!Tk_IsMapped(tkwin)) {
	return;
    }
    
    tw = Tk_Width(tkwin);
    th = Tk_Height(tkwin);
    
    // Create pixmap for double-buffering if necessary
    if(tmln->doubleBuffer)  {
	pm = XCreatePixmap(disp, Tk_WindowId(tkwin), tw, th, 
			   DefaultDepthOfScreen(Tk_Screen(tkwin)));
	d = pm;
    } else {			// draw directly into window
	d = Tk_WindowId(tkwin);
    }
    // redraw background
    Tk_Fill3DRectangle(tkwin,d,tmln->bgBorder,0,0, tw, th, tmln->borderWidth, 
		       tmln->relief);
    // Actually do the drawing
    inset = 2;
    
    //    drawX(tkwin, tmln->gc, d, tmln->borderWidth + inset);
    VUMeterDraw(tmln, d);
    
    // copy to screen if double-buffering
    if(tmln->doubleBuffer)  {
	XCopyArea(disp, pm, Tk_WindowId(tkwin), tmln->gc, 0, 0, 
		  tw, th, 0, 0);
	XFreePixmap(disp, pm);
    }
}

static void VUMeterDestroy(char* clientData)
{				/* destroy one of our objects */
    VUMeter *tmln = (VUMeter *)clientData;
    
    Tk_FreeOptions(VUMeterConfigSpecs, (char *)tmln, tmln->display, 0);
    if(tmln->gc != None)  {
	Tk_FreeGC(tmln->display, tmln->gc);
    }
    // don't destroy sound as we didn't new it
    ckfree((char *)tmln);
}


