#!/usr/local/bin/wish
# -*- Tcl -*-
#
# audio_frontend.tcl
#
# Interactive front-end to sample speech from a local microphone.
# Installed version.
#
# dpwe@icsi.berkeley.edu, after graphic.tcl. 1995feb23
# $Header: /u/drspeech/src/audio_frontend/RCS/audio_frontend.tcl,v 1.26 1999/05/06 02:09:45 dpwe Exp $

# Capture the RCS version for display
set afe_revision {$Revision: 1.26 $}

# Do we want to see the 'buffer-full' meter?
set usefill 0
# Show the voice-detect lamp indicator?
set useindic 0
# Show the 'slow' VU meter (in addition to the fast one)?
set useslowvu 0
# Do we want the load/save/resubmit buttons (e.g. for debugging bad sounds)
set useloadsave 0
# Do we want the quit button?
set usequit 1

# Do we want to display results fed from a port
set resultport ""

# How far above floor to be speech? (dB)
set hithresh 15.0

# Shortest speech file we will accept (secs)
set minRecordTime 1.2
# interacts strongly with lowpass cutoff of vulpf, since this extends all snds

# Sound data characteristics
set samplerate 16000

# show debug messages?
set debug 0

# You can specify $playcmd as an external program to use when playing 
# back recorded files.  If it's empty, we will create dynamic APRL 
# sinks to do playback, but that's buggy on some older versions of 
# solaris.  As a special case, "playcmd=1" invokes the default 
# play command, $SPEECH_BIN/sndplay .  See below for this logic
set playcmd ""

# Do we filter out noway leading numbers from messages over the resultport?
set nowaymsgs 1

# What soundfile format should the output stream have?
# This must be a valid argument to the -T flag of sndrfmt (1)
set outputfmt "PCM/Abb"

# Modifications for embedding in other Tcl programs:
# (1) If $afe_root is set on entry, all widgets will be built below 
#     this root widget.
# (2) If $afe_outputHandler is set on entry, then the output 
#     of the processing chain is passed to a Tcl pipe with this 
#     routine as the fileevent handler (with the pipe name appended).
#     (Normally, the pipe output is just written to stdout)
# (3) $afe_soundfile is set to the name of the temporary file written 
#     containing the entire utterance.
# $afe_outputHook will be executed just before we sent speech onward
# afe_setstatus can be called to write into the status window

###################### no obvious customization below here #############

# Load our common bits from this directory
source "[file dirname [info script]]/afe_utils.tcl"

# Make sure APRL is loaded if this is plain wish
# (using Tcl 7.5's 'package' system)
if {[catch {package require "APRL"} rslt]} {
    puts stderr "**$argv0: can't find required APRL package"
    puts stderr "**$argv0: - contact dpwe@icsi.berkeley.edu"
    puts stderr "**$argv0: (result was: $rslt)"
    puts stderr "**$argv0: (auto_path was: $auto_path)"
    puts stderr "**$argv0: (tcl_pkgPath was: $tcl_pkgPath)"
    exit -1
}

# Default values for the pipe components
set featureprog ""
set featureargs ""
set onlineprog ""
set onlineargs ""
set fwpprog ""
set fwpargs ""

# Parse command-line args=val options
GetICSIargs

# dummy proc for auto-loading
proc audio_frontend_init {} {
}

# for the fileevent: we need central management of fileevent readable
# listeners in integ_frontend;  we do this by setting them up with 
# a call to "rdprocAdd $pipe $cmd" instead of "fileevent $pipe readable $cmd"
# But if we're not embedded, we want to define our own rdprocAdd to do 
# the default thing.
if {[info proc rdprocAdd] == ""} {
    proc rdprocAdd {pipe cmd} {
	# Dummy for when rdprocmgr is not invoked
	fileevent $pipe readable $cmd
    }
}

# Make sure required external programs exist
set sndrfmt "sndrfmt"
if {[catch {exec $sndrfmt} rslt] && [string match "*no such*" $rslt]} {
    puts stderr "**$argv0: can't find executable $sndrfmt"
    puts stderr "**$argv0: - needed to output - contact dpwe@icsi.berkeley.edu"
    puts stderr "**$argv0: (result was: $rslt)"
    puts stderr "**$argv0: (path was: $env(PATH))"
    exit -1
}

if {$playcmd == "1"} {
    # special case to use a default external play command
    set playcmd "sndplay"
}
if {$playcmd != "" \
	&& [catch {exec $playcmd} rslt] && [string match "*no such*" $rslt]} {
    puts stderr "**$argv0: can't find executable $playcmd"
    puts stderr "**$argv0: - needed to play - contact dpwe@icsi.berkeley.edu"
    puts stderr "**$argv0: (result was: $rslt)"
    puts stderr "**$argv0: (path was: $env(PATH))"
    exit -1
}

if {[info var afe_outputHandler] == ""} {
    # No handler has been specified, so we're writing to stdout.
    set afe_outputHandler ""
    # Make sure there's no hidden EOL translation for binary output stdout 
    # (Tcl 7.5 +)
    fconfigure stdout -translation binary -buffering none
} else {
    # An embedding program specified where we should send the output
}

if {[info var afe_outputHook] == ""} {
    # Set a default hook
    set afe_outputHook ""
}

set tcl_interactive 0
#wm minsize . 1 1

# Name of the file we will write to (temporary)
#set soundfilename "tmp.aif"
# This variable may be called upon by embedding code
set afe_soundfile "/tmp/frontend.speech.[pid].aif"

### Ok, now start building the sigproc network...

# initialize the scheduler
scheduler sched -timestep 0.1

# source for playing back recorded files
set playsrc [soundfsrc ps]

# standard input process
set afe_audiosrc [audiosrc as -sampleRate $samplerate -bufferTime 0.1]

# Default output process is none, unless play is started
set audiosink ""
#audiosink ak -sampleRate $samplerate -scale 0.0001

# run an rms for the VU meter
hpfilter hp      -input as -cutoff 500
rms      vurms   -sampleRate 100
lpfilter vulpf2  -cutoff 20
dB	 vudb2

# delay line for getting the sound from a short while ago (was 0.2 s)
set delaytime 0.4
set postdelaytime $delaytime
delay    dline   -input $afe_audiosrc -delayTime $delaytime
# store this as the widget which we record from
set soundsource "dline"

#delay    vudline -input vurms     -delayTime $delaytime -initVal 0.00001
lpfilter vulpf   -cutoff 5 -input vurms
dB       vudb    

# flag that playsink exists
set state_playing 0
# record the current state of whether we think speech is present or not
set speechon 0

proc recalibrate {} {
    # Trigger a re-estimation of the noise floor, after getting confused
    global floor
    set floor 50.0
    DEBUG "recalibrating..."
}

set afe_cont 1
# standard execution code
proc cont {} {
    global state_playing playsrc afe_audiosrc audiosink speechon indic floor
    global usefill delaytime useslowvu afe_root afe_cont
    sched step
    # Update the VU meters
    set slowval [vudb val]
    set quikval [vudb2 val]
    $afe_root.vu config -val $quikval
    if {$useslowvu} {
	$afe_root.vu2 config -val $slowval
    }
    if {[sched cget -nowtime]>($delaytime+0.2)} {
	# Run the speech detection
	if {$slowval < ($floor-0.5)} {
	    # Only take new value if it's significantly lower than current
	    # so changes are notified relatively rarely
	    set floor $slowval
	    DEBUG "new floor: $floor dB"
	}
	# How far above floor to be speech?
#	set hithresh 20.0
	global hithresh
	set lothresh [expr $hithresh-5.0]
	if {$quikval > $floor+$hithresh} {
	    if {$speechon == 0} {
		set speechon 1
		$indic select
		ev_SpeechOn
	    }
	} else {
	    # Quickval is not peaking, so we may consider the slowval
	    if {$speechon != 0 && $slowval < $floor+$lothresh} {
		set speechon 0
		$indic deselect
		ev_SpeechOff
	    }
	}
    }

    # Update input buffer fillmeter
    if {$usefill} {
	$afe_root.fill config -val [as filled]
    }
#    # Quit if we've run out of input (for timing tests anyway)
#    if { ![as more] } { exit }
    # end 'play' if it was active
    if { $audiosink != "" && [$playsrc more]==0 } {
	# Playback was active, but it ended
	afe_playdone
	if {$state_playing} {
	    ev_playDone
	}
    }
    # Arrange for the next execution of this update routine
    if {$afe_cont} {after 10 { cont }}
}

## WIDGET CREATION
#
# VU meter on the left, buffer monitor on the bottom, buttons on the right, 
# potential configuration panel on the far right

# If we're passed in $afe_root, use it as the basis for our widgets
# else make it nothing
if {[info vars "afe_root"] == ""} {
    set afe_root ""
}

#frame $afe_root.config 
# skip the config for now...

############ Result panel ################
if {$resultport != ""} {
    frame $afe_root.rslt -relief groove -borderwidth 2
    label $afe_root.rslt.label -text "Result:"
    set rsltwin [entry $afe_root.rslt.rslt -relief sunken -width 32]
    pack $afe_root.rslt.label -side top -fill y    -expand 0
    pack $afe_root.rslt.rslt  -side top -fill both -expand 1
    pack $afe_root.rslt -side bottom -fill x -expand 1

    proc SetResult {msg {color gray85}} {
	# Display a message in the result window
	global rsltwin
	$rsltwin delete 0 e
	$rsltwin insert 0 $msg
	$rsltwin config -background $color
    }

    proc ResultPortCmd {chan cliAdr cliPort} {
	# Handle a port opening
	DEBUG "ResultPortCmd: chan=$chan cliAdr=$cliAdr cliPort=$cliPort"
	set msg [gets $chan]
	set color gray85
	global nowaymsgs
	if {$nowaymsgs} {
	    set num 0
	    # extract the integer leading the string
	    regexp {^[0-9]+} $msg num
	    # .. and remove it
	    regsub {^[0-9]+ *} $msg "" msg
	    # Filter out the noway startup message while we're at it
	    regsub "^noway version.*" $msg "" msg
	    if {$num > 0 && $num < 20} {
		# This number suggests a final hypothesis - color it different
		set color white
	    }
	}
	# Don't display the message if we've filtered it down to nothing
	if {$msg != ""} {
	    SetResult $msg $color
	}
	close $chan
    }

    socket -server ResultPortCmd $resultport
}

## vu meter to show input buffer fillation
# max scale depends on input buffer size
set inbufsize 2000
if {$usefill} {
 vumeter $afe_root.fill -scale [expr 1.0/$inbufsize] -relief flat -height 10 -border 3 \
	-padX 0 -padY 0 -cells 100 -nyellow 0 -ngreen 100 -green red \
	-frame 0 -fg #cdb79e -width 100
 pack $afe_root.fill -side bottom -expand yes -fill x
}

# VU meter on left-hand edge (displays from -50 to 0 (dB))
vumeter $afe_root.vu -width 32 -orient vert -relief sunken -scale 0.02 -offset 1
pack $afe_root.vu -side left -fill y -expand 0 -padx 5 -pady 5
## 2nd vu meter
if {$useslowvu} {
    vumeter $afe_root.vu2 -width 32 -orient vert -relief sunken -scale 0.02 -offset 1
    pack $afe_root.vu2 -side left -fill y -expand 0 -padx 5 -pady 5
}

# Button box
frame $afe_root.buttons
# Make the buttons and display them
set buttons {}
lappend buttons [set recordbutton [button $afe_root.buttons.recd \
	-text "Record speech" \
	-command "ev_recordPress"]]
lappend buttons [set stopbutton [button $afe_root.buttons.stop \
	-text "Stop recording" -state disabled \
	-command "ev_stopPress"]]
lappend buttons [set playbutton [button $afe_root.buttons.play -text "Play speech" \
	-command "ev_playPress" -state disabled]]
if {$useloadsave} {
    lappend buttons [set loadbutton [button $afe_root.buttons.load -text "Load speech ..." -command LoadFile]]
    lappend buttons [set savebutton [button $afe_root.buttons.save -text "Save speech ..." -command SaveFile -state disabled]]
    lappend buttons [set fakebutton [button $afe_root.buttons.resend -text "Resubmit speech" \
	                   -command sendSpeechOnward -state disabled]]
}
#lappend buttons [button $afe_root.buttons.opts -text "Show options" -command showopts]
if {$usequit} {
    lappend buttons [set quitbutton [button $afe_root.buttons.quit -text "Quit"  -command Exit]]
}
lappend buttons [set indic [checkbutton $afe_root.buttons.indi -indicatoron 1 -text ""]]
lappend buttons [frame  $afe_root.buttons.status]
# Bottom of button frame is status label/text pair
label $afe_root.buttons.status.label -text "Status:"
set statuswin [entry $afe_root.buttons.status.status -relief sunken -width 16]
pack $afe_root.buttons.status.label -side left -fill y -expand 0
pack $afe_root.buttons.status.status -side left -fill both -expand 1

# Note the AFE revision small on the screen
regexp {:(.*)\$} $afe_revision all revision
label $afe_root.name -text "audio_frontend $revision" -font 5x8
pack $afe_root.name -side top -fill x -expand 1

eval "pack $buttons  -side top -fill x -expand 0 -padx 2 -pady 2"
pack $afe_root.buttons -side top -fill both -expand 1
# not the indicator
if {!$useindic} {
    pack forget $indic
}

proc Exit {{code 0}} {
    # cleanup & exit
    global afe_soundfile
    if {[file exist $afe_soundfile]} {
	catch "exec /bin/rm $afe_soundfile"
    }
    exit $code
}

proc showopts {} {
    # Popup the panel of options that may be set
    puts stderr "*** SHOWOPTS NOT YET IMPLEMENTED ***"
}

proc LoadFile {} {
    # Load button pressed
    global afe_soundfile samplerate sndrfmt
    set infile [tk_getOpenFile]
    if {$infile != ""} {
	catch {exec $sndrfmt -TAIFF $infile $afe_soundfile} rslt
	if {[string match "*could not recog*" $rslt]} {
	    # Treat as headerless data
	    catch {exec $sndrfmt -TAIFF -a -r $samplerate $infile $afe_soundfile} rslt
	}
	puts stderr $rslt
	# Really, need to check sampling rate...
    }
    # Make sure buttons are all correct
    mcState "idle"
}

proc SaveFile {} {
    # Save button pressed
    global afe_soundfile samplerate sndrfmt
    set outfile [tk_getSaveFile]
    if {$outfile != ""} {
	# Save it as raw data, or as an AIFF?
	#catch {exec $sndrfmt -s $afe_soundfile $outfile} rslt
	catch {exec $sndrfmt -S PCM/R${samplerate} -TAIFF $afe_soundfile $outfile} rslt
	puts stderr $rslt
    }
    # Make sure buttons are all correct
    mcState "idle"
}

proc sendSpeechOnward {} {
    # pass the data onward to the output
    # First, make sure that the file has finished being written
    record_wait_for_done
    DEBUG "Real speech captured NOW"
    global afe_soundfile samplerate sndrfmt outputfmt afe_outputHook
    if {$afe_outputHook != ""} {
	eval $afe_outputHook
    }
    set ecmd "$sndrfmt -T $outputfmt $afe_soundfile -"
    global featureprog featureargs onlineprog onlineargs fwpprog fwpargs
    if {$featureprog != ""} {
	append ecmd " | $featureprog $featureargs"
    }
    if {$onlineprog != ""} {
	append ecmd " | $onlineprog $onlineargs"
    }
    if {$fwpprog != ""} {
	append ecmd " | $fwpprog $fwpargs"
    }
    global argv0
    DEBUG "$argv0: about to execute '$ecmd'"
    global afe_outputHandler
    if {$afe_outputHandler != ""} {
	# arrange for the command output to come back to Tcl
	set speechPipe [open "| $ecmd" "r"]
	fconfigure $speechPipe -translation binary -buffering none -blocking 0
	#fileevent $speechPipe readable "$afe_outputHandler $speechPipe"
	rdprocAdd $speechPipe "$afe_outputHandler $speechPipe"
    } else {
	# Just send output to stdout
	eval "exec $ecmd &"
    }
    # Clear the result window
    global resultport
    if {$resultport != ""} {
	SetResult "(waiting)"
    }
}

proc DEBUG msg {
    # just report debugging messages to the console
    global debug
    if {$debug} {
	puts stderr "DEBUG: $msg"
    }
}

proc ev_recordPress {} {
    # "Record" button pressed
    global afe_state speechon
    switch $afe_state {
	"idle" {
	    # was a "record" command - move to the "listen" state
	    if {$speechon} {
		# .. unless there is already speech, then just record
		mc_toRecord
	    } else {
		mc_toListen
	    }
	    # If we have a result panel, clear it
	    global resultport
	    if {$resultport != ""} {
		SetResult ""
	    }
	}
    }
}

proc ev_stopPress {} {
    # "Stop recording" button pressed
    global afe_state speechon
    switch $afe_state {
	"recording" {
	    # was a "stop recording command - 
	    # also, recalibrate to perhaps avoid the problem that made 
	    # us want to stop recording manually
	    recalibrate
	    mc_fromRecord
	}
	"listening" {
	    # "stop recording" when just "listening" - go straight to idle
	    # also, recalibrate to perhaps avoid the problem that made 
	    # us want to stop recording manually
	    recalibrate
	    mc_toIdle
	}
    }
}

set afe_noRecordDuringPlay 1

proc afe_playfile {filename} {
    # Play a file through the APRL modules
    global playcmd playsrc samplerate afe_audiosrc audiosink
    global afe_noRecordDuringPlay
    # Not already playing
    # Maybe stop recording for non-duplex hardware
    if {$afe_noRecordDuringPlay} {
	$afe_audiosrc close
    }
    if {$audiosink == ""} {
	if {$playcmd != ""} {
	    # just run an external command
	    catch "exec $playcmd $filename"
	    # but set the flag to call afe_playdone when it's done
	    set audiosink "1"
	} else {
	    # Stay within APRL
	    $playsrc config -filename $filename
	    set audiosink [audiosink ak -sampleRate $samplerate -input $playsrc -scale 1.0]
	}
    }
    return $audiosink
}

proc afe_playdone {} {
    # Clean up after play
    global afe_noRecordDuringPlay afe_audiosrc audiosink
    if {$audiosink != "" && $audiosink != "1"} {
	# Destroy the sink
	rename $audiosink {}
    }
    # Clear this semaphore
    set audiosink ""
    # Maybe re-enable input
    if {$afe_noRecordDuringPlay} {
	$afe_audiosrc open
    }
}

proc ev_playPress {} {
    # Play back the recorded file
    # Can do this within aprl, or with an external command.
    global playbutton state_playing afe_soundfile
    if {$state_playing == 0} {
	$playbutton config -state disabled
	afe_playfile $afe_soundfile
	set state_playing 1
    }
}

proc ev_playDone {} {
    # Mainloop detected the end of playback
    global state_playing playbutton
    # re-enable the button
    $playbutton config -state normal
    # Finish up
    #afe_playdone
    # no - called independently now
    # clear the semaphor
    set state_playing 0
}

proc ev_SpeechOn {} {
    # event is: speech signal detected
    global afe_state
    switch $afe_state {
	"listening" {
	    mc_toRecord
	}
    }
}

proc ev_SpeechOff {} {
    # event is: speech signal lost
    global afe_state
    switch $afe_state {
	"recording" {
	    mc_fromRecord
	}
    }
}

# State machine for speech recording
set afe_state "idle"
# idle -> listening -> recording (->listening) -> idle

# External routine for writing to the status window
proc afe_setstatus state {
    global statuswin
    $statuswin delete 0 e
    $statuswin insert 0 $state
    if {$state == "recording"} {
	$statuswin config -background red
    } elseif {$state == "listening"} {
	$statuswin config -background palegreen3
    } elseif {$state == "idle"} {
	$statuswin config -background gray85
    } else {
	# Some other message supplied by application - use pale blue
	$statuswin config -background lightsteelblue1
    }
}

proc mcState state {
    # set the state to $state & handle all the user-interface updates
    global statuswin playbutton recordbutton stopbutton afe_state 
    global afe_soundfile
    global useloadsave loadbutton savebutton fakebutton
    afe_setstatus $state
    if {$state == "recording" || $state == "listening"} {
	# not idle - listening or recording
	$recordbutton config -state disabled
	$stopbutton config -state normal
	$playbutton config -state disabled
	if {$useloadsave} {
	    $savebutton config -state disabled
	    $loadbutton config -state disabled
	    $fakebutton config -state disabled
	}
    } else {
	# presumably "idle"
	$recordbutton config -state normal
	$stopbutton config -state disabled
	if {$useloadsave} {
	    $loadbutton config -state normal
	}
	# If there's a soundfile lying around, allow it to be played
	if {[file readable $afe_soundfile]} {
	    $playbutton config -state normal
	    if {$useloadsave} {
		$savebutton config -state normal
		$fakebutton config -state normal
	    }
	}
    }
    # Make sure display is updated promptly
    update
    DEBUG "state: $state"
    set afe_state $state
}

proc mc_toListen {} {
    # move to listening state (from idle)
    mcState "listening"
}

proc mc_toIdle {} {
    # move to idle state (from listening(?))
    mcState "idle"
}

proc mc_toRecord {} {
    # entering the recording state (from listening)
    record_open
    mcState "recording"
}

# a way to pick up how much time actually was recorded (for embedding)
set afe_recordTime ""

proc mc_fromRecord {} {
    # Leave the recording state; may go back to listening or to idle
    global minRecordTime
    global afe_recordTime
    set recordTime [record_close]
    if {$recordTime < $minRecordTime} {
	# We didn't get enough speech - go back to listening
	mcState "listening"
    } else {
	# We got enough speech to be worth processing, so revert to idle
	mcState "idle"
	# Record how long this speech was
	set afe_recordTime $recordTime
	# .. and pass the speech we captured on to later processing
	sendSpeechOnward
    }
}

# Keep track of when recording started, to figure how much we have caught
set recordStartTime 0
# A flag to say if we're currently writing output to a file (asynch)
set recordSemaphore 0

proc record_open {} {
    # Begin writing the audio input to a file
    # Make sure we're not already writing
    record_wait_for_done
    global afe_soundfile soundsource
    global recordStartTime recordSemaphore
    set recordStartTime [sched cget -nowtime]
    # start the actual recording
    # create a sound file sink, attach to the named pipe 
    soundfsink sfsink -filename $afe_soundfile -input $soundsource
    # set semaphore to say we are running
    set recordSemaphore 2
}

proc record_close {} {
    # cease recording the audio input
    # Return the total time in seconds recorded
    global postdelaytime
    global recordStartTime recordSemaphore
    # set semaphore to say we are closing
    set recordSemaphore 1
    # how long we have been recording?
    set gottime [expr [sched cget -nowtime] - $recordStartTime]
    # OK, arrange for the file to be closed soon
    # Add a little tail-end time (400 ms) to make sure we get everything
    after [expr int(1000*$postdelaytime)] record_close_doit
    DEBUG "got: $gottime secs"
    return $gottime
}

proc record_close_doit {} {
    # Actually close the file
    global recordSemaphore
    # Close the soundfile if it exists (sometimes apparently doesn't???)
    if {[info commands sfsink] != ""} {
	sfsink close
	# Destroy the soundfilesink object
	rename sfsink {}
    } else {
	puts stderr "record_close_doit: no sfsink command found"
    }
    # Indicate that the file is free
    set recordSemaphore 0
}

proc record_wait_for_done {} {
    global recordSemaphore
    if {$recordSemaphore } {
	if {$recordSemaphore > 1} {
	    puts stderr "record_wait_for_done: not about to stop!?"
	}
	# if it's 1, still being written - wait for asynchronous close
	tkwait variable recordSemaphore
	# Assume, if it changed, it went to 0
    }
}

# Initialize the noise floor
recalibrate
# Start it all off
mcState idle
cont

