#!/bin/sh
# -*- Tcl -*-   The next line restarts using drspeech_tclsh \
	exec drspeech_tclsh $0 ${1+"$@"}

#!/usr/local/bin/tclsh
#
# test-bfcp.tcl
#
# Test the bfilecopy fn (after Malcolm found a recondite bug).
#
# 1997nov06 dpwe@icsi.berkeley.edu
# $Header: /u/drspeech/src/SPRACHcore/guitools/dpweutils_tcl/RCS/test-bfcp.tcl,v 1.2 1998/12/02 23:59:20 dpwe Exp $

#lappend auto_path "/u/dpwe/lib/$env(MACHTYPE)"
#package require Dpweutils_tcl

# Either we're given the name of the lib to load as an arg, or 
# we guess it's in the standard place relative to us
if {$argv != ""} {
    set libname $argv
} else {
    set libname "[file dirname [info script]]/H-$env(MACHTYPE)/libdpweutils_tcl.so"
}

if {![string match "*/*" $libname]} {
    # No path components - prepend cwd
    set libname "[pwd]/$libname"
}
puts "loading $libname ..."
load $libname

# Basically, if we get a partial match, because all the characters in 
# the buffer look plausible, we then read enough more to check the 
# rest, but if that check fails, we don't check for the stop string 
# in the rest of the buffer.

proc test_bfilecopy {prefix stopstring} {
    # Run a test case with bfilecopy using the specified stop string, 
    # and inserting the given prefix just to check it out

    # Construct the test file
    set test "test.data"
    set out1 "out.data"

    set tf [open $test "w"]
    puts $tf "line 1"
    puts $tf "line 2"
    # This one worked OK even on the old version because the 
    # blocking meant that the decoy prefix was clearly wrong even on the 
    # first pass (i.e. it read int "stopsto" which isn't a prefix of
    # "stopstring").
    puts $tf "      "
    puts -nonewline $tf $prefix
    # Write out a decoy prefix of the stopstring
    puts -nonewline $tf [string range $stopstring 0 3]
    # The first part should be read as everything including the decoy
    set part1len [tell $tf]
    # Write out the separator
    puts -nonewline $tf "$stopstring"
    # Figure the length of the second part
    set part2base [tell $tf]
    puts $tf "\nnext bit"
    set part2len [expr [tell $tf] - $part2base]
    close $tf

    set instream [open $test "r"]

    set outstream [open $out1 "w"]
    set len1 [bfilecopy $instream $outstream -1 $stopstring]
    close $outstream

    set if [open $out1 "r"]
    set got1 [read $if]
    close $if

    set outstream [open $out1 "w"]
    set len2 [bfilecopy $instream $outstream -1 $stopstring]
    close $outstream

    set if [open $out1 "r"]
    set got2 [read $if]
    close $if

    close $instream

    if {$len1 != $part1len || $len2 != $part2len} {
	puts "got1='$got1'"
	puts "got2='$got2'"

	puts "len1=$len1 len2=$len2"
	puts "part1len=$part1len part2len=$part2len"

	puts "FAILED - didn't find stop string '$stopstring' after prefix '$prefix'"
	exit -1
    }
    file delete $test
    file delete $out1
}

set prefix "123456789ABCDE"
foreach stopstring "stopstring -------*" {
    for {set i -1} {$i < [string length $prefix]} {incr i} {
	test_bfilecopy [string range $prefix 0 $i] $stopstring
    }
}

puts stderr "bfilecopy tests passed OK"
